# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np


# specific
from . import _generic_check


_ELEMENTS = 'knots'


# #############################################################################
# #############################################################################
#                           Mesh2DRect
# #############################################################################


def _mesh2DRect_X_check(
    x=None,
    res=None,
):
    """ Returns knots (x) and associated resolution

    res can be:
        - int: numbr of mesh elements desired between knots
        - float: desired average mesh element size
        - array of floats: (one for each x, desired approximate mesh size)

    """

    # ------------
    # Check inputs

    # x
    try:
        x = np.unique(np.ravel(x).astype(float))
    except Exception as err:
        msg = "x must be convertible to a sorted, flat array of floats!"
        raise Exception(msg)

    # res
    if res is None:
        res = 10

    lc = [
        isinstance(res, (int, np.int64, np.int32)) and len(x) == 2,
        isinstance(res, (float, np.floating)) and len(x) == 2,
        isinstance(res, (list, tuple, np.ndarray)) and len(x) == len(res),
    ]
    if not any(lc):
        msg = (
            "Arg res must be:\n"
            "\t- int: nb of mesh elements along x\n"
            "\t       requires len(x) = 2\n"
            "\t- float: approximate desired mesh element size along x\n"
            "\t       requires len(x) = 2\n"
            "\t- iterable: approximate desired mesh element sizes along x\n"
            "\t       requires len(x) = len(res)\n"
        )
        raise Exception(msg)

    if lc[0]:
        x_new = np.linspace(x[0], x[1], int(res)+1)
        res_new = res
        indsep = None

    elif lc[1]:
        nb = int(np.ceil((x[1]-x[0]) / res))
        x_new = np.linspace(x[0], x[1], nb+1)
        res_new = np.mean(np.diff(x))
        indsep = None

    else:

        # check conformity
        res = np.ravel(res).astype(float)
        delta = np.diff(x)
        res_sum = res[:-1] + res[1:]
        ind = res_sum > delta + 1.e-14
        if np.any(ind):
            msg = (
                "Desired resolution is not achievable for the following:\n"
                f"res_sum: {res_sum[ind]}\n"
                f"delta  : {delta[ind]}"
            )
            raise Exception(msg)

        # compute nn
        # nn = how many pairs can fit in the interval
        npairs = np.round(delta/res_sum).astype(int)
        res_sum_new = delta / npairs

        fract = res[:-1] / res_sum

        res_new = [None for ii in range(len(x)-1)]
        x_new = [None for ii in range(len(x)-1)]
        for ii in range(len(x)-1):
            res_new[ii] = (
                res_sum_new[ii]
                * np.linspace(fract[ii], 1.-fract[ii], 2*npairs[ii])
            )
            if ii == 0:
                res_add = np.concatenate(([0], np.cumsum(res_new[ii])))
            else:
                res_add = np.cumsum(res_new[ii])
            x_new[ii] = x[ii] + res_add

        indsep = np.cumsum(npairs[:-1]*2)
        res_new = np.concatenate(res_new)
        x_new = np.concatenate(x_new)

    return x_new, res_new, indsep


def _mesh2DRect_check(
    domain=None,
    res=None,
):

    # --------------
    # check inputs

    # domain
    c0 = (
        isinstance(domain, list)
        and len(domain) == 2
        and all([hasattr(dd, '__iter__') and len(dd) >= 2 for dd in domain])
    )
    if not c0:
        msg = (
            "Arg domain must be a list of 2 iterables of len() >= 2\n"
            f"Provided: {domain}"
        )
        raise Exception(msg)

    # res
    c0 = (
        res is None
        or np.isscalar(res)
        or isinstance(res, list) and len(res) == 2
    )
    if not c0:
        msg = (
            "Arg res must be a int, float or array or a list of 2 such\n"
            f"Provided: {res}"
        )
        raise Exception(msg)

    if np.isscalar(res) or res is None:
        res = [res, res]

    # -------------
    # check R and Z

    R, resR, indR = _mesh2DRect_X_check(domain[0], res=res[0])
    Z, resZ, indZ = _mesh2DRect_X_check(domain[1], res=res[1])

    return R, Z, resR, resZ, indR, indZ


def _mesh2DRect_to_dict(
    domain=None,
    res=None,
    key=None,
):

    # --------------------
    # check / format input

    if not isinstance(key, str):
        msg = "Arg key must be a str!"
        raise Exception(msg)

    kRknots, kZknots = f"{key}-R-knots", f"{key}-Z-knots"
    kRcent, kZcent = f"{key}-R-cents", f"{key}-Z-cents"

    R, Z, resR, resZ, indR, indZ = _mesh2DRect_check(domain=domain, res=res)
    Rcent = 0.5*(R[1:] + R[:-1])
    Zcent = 0.5*(Z[1:] + Z[:-1])

    variable = not (np.isscalar(resR) and np.isscalar(resZ))

    # --------------------
    # prepare dict

    dref = {
        kRknots: {
            'data': R,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'R',
            'name': 'R',
            'group': 'R',
        },
        kZknots: {
            'data': Z,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'group': 'Z',
        },
        kRcent: {
            'data': Rcent,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'R',
            'name': 'R',
            'group': 'R',
        },
        kZcent: {
            'data': Zcent,
            'units': 'm',
            # 'source': None,
            'dim': 'distance',
            'quant': 'Z',
            'name': 'Z',
            'group': 'Z',
        },
    }

    # dobj
    dmesh = {
        key: {
            'type': 'rect',
            'knots': (kRknots, kZknots),
            'cents': (kRcent, kZcent),
            'ref': (kRcent, kZcent),
            'shape': (Rcent.size, Zcent.size),
            'variable': variable,
            'crop': False,
        },
    }
    return dref, dmesh


def _mesh2DRect_from_Config(config=None, key_struct=None):

    # ------------
    # check inputs

    if not config.__class__.__name__ == 'Config':
        msg = "Arg config must be a Config instance!"
        raise Exception(msg)

    # -------------
    # key_struct if None

    if key_struct is None:
        lk, ls = zip(*[
            (ss.Id.Name, ss.dgeom['Surf']) for ss in config.lStructIn
        ])
        key_struct = lk[np.argmin(ls)]

    # -------------
    # domain

    poly = config.dStruct['dObj']['Ves'][key_struct].Poly_closed
    domain = [
        [poly[0, :].min(), poly[0, :].max()],
        [poly[1, :].min(), poly[1, :].max()],
    ]
    return domain, poly


# #############################################################################
# #############################################################################
#                           Mesh2DRect - select
# #############################################################################


def _select_ind_check(
    ind=None,
    elements=None,
    returnas=None,
    crop=None,
):

    # ind
    lc = [
        ind is None,
        isinstance(ind, tuple)
        and len(ind) == 2
        and (
            all([np.isscalar(ss) for ss in ind])
            or all([
                hasattr(ss, '__iter__')
                and len(ss) == len(ind[0])
                for ss in ind
            ])
            or all([isinstance(ss, np.ndarray) for ss in ind])
        ),
        (
            np.isscalar(ind)
            or (
                hasattr(ind, '__iter__')
                and all([np.isscalar(ss) for ss in ind])
            )
            or isinstance(ind, np.ndarray)
        )
    ]

    if not any(lc):
        msg = (
            "Arg ind must be either:\n"
            "\t- None\n"
            "\t- int or array of int: int indices in mixed (R, Z) indexing\n"
            "\t- tuple of such: int indices in (R, Z) indexing respectively\n"
            f"Provided: {ind}"
        )
        raise Exception(msg)

    if lc[0]:
        pass
    elif lc[1]:
        if any([not isinstance(ss, np.ndarray) for ss in ind]):
            ind = (
                np.atleast_1d(ind[0]).astype(int),
                np.atleast_1d(ind[1]).astype(int),
            )
        lc0 = [
            [
                isinstance(ss, np.ndarray),
                np.issubdtype(ss.dtype, np.integer),
                ss.shape == ind[0].shape,
            ]
                for ss in ind
        ]
        if not all([all(cc) for cc in lc0]):
            ltype = [type(ss) for ss in ind]
            ltypes = [
                ss.dtype if isinstance(ss, np.ndarray) else False
                for ss in ind
            ]
            lshapes = [
                ss.shape if isinstance(ss, np.ndarray) else len(ss)
                for ss in ind
            ]
            msg = (
                "Arg ind must be a tuple of 2 arrays of int of same shape\n"
                f"\t- lc0: {lc0}\n"
                f"\t- types: {ltype}\n"
                f"\t- type each: {ltypes}\n"
                f"\t- shape: {lshapes}\n"
                f"\t- ind: {ind}"
            )
            raise Exception(msg)
    else:
        if not isinstance(ind, np.ndarray):
            ind = np.atleast_1d(ind).astype(int)
        c0 = (
            np.issubdtype(ind.dtype, np.integer)
            or np.issubdtype(ind.dtype, np.bool_)
        )
        if not c0:
            msg = (
                "Arg ind must be an array of bool or int\n"
                f"Provided: {ind.dtype}"
            )
            raise Exception(msg)

    # elements
    elements = _generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    returnas = _generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default=tuple,
        allowed=[tuple, np.ndarray, 'tuple-flat', 'array-flat', bool],
    )

    # crop
    crop = _generic_check._check_var(
        crop, 'crop',
        types=bool,
        default=True,
    )

    return ind, elements, returnas, crop


def _select_check(
    elements=None,
    returnas=None,
    return_neighbours=None,
):

    # elements
    elements = _generic_check._check_var(
        elements, 'elements',
        types=str,
        default=_ELEMENTS,
        allowed=['knots', 'cents'],
    )

    # returnas
    returnas = _generic_check._check_var(
        returnas, 'returnas',
        types=None,
        default='ind',
        allowed=['ind', 'data'],
    )

    # return_neighbours
    return_neighbours = _generic_check._check_var(
        return_neighbours, 'return_neighbours',
        types=bool,
        default=True,
    )

    return elements, returnas, return_neighbours,


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def _mesh2DRect_bsplines(key=None, lkeys=None, deg=None):

    # key
    if key is None and len(lkeys) == 1:
        key = lkeys[0]
    if key not in lkeys:
        msg = (
            "Arg key must be a valid mesh identifier!\n"
            f"\t- available: {lkeys}\n"
            f"\t- provided: {key}"
        )
        raise Exception(msg)

    # deg
    if deg is None:
        deg = 2
    if not isinstance(deg, int) and deg in [0, 1, 2, 3]:
        msg = (
            "Arg deg must be a int in [0, 1, 2, 3]\n"
            f"Provided: {deg}"
        )
        raise Exception(msg)

    # keybs
    keybs = f'{key}-bs{deg}'

    return key, keybs, deg





import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import warnings


from ._DataCollection_class import DataCollection
from . import _comp_spectrallines
from . import _DataCollection_comp
from . import _DataCollection_plot


__all__ = ['SpectralLines', 'TimeTraces']


_OPENADAS_ONLINE = True

_GROUP_LINES = 'lines'
_GROUP_NE = 'ne'
_GROUP_TE = 'Te'
_UNITS_LAMBDA0 = 'm'


#############################################
#############################################
#       Spectral Lines
#############################################


class SpectralLines(DataCollection):

    _ddef = {
        'Id': {'include': ['Mod', 'Cls', 'Name', 'version']},
        'params': {
            'lambda0': (float, 0.),
            'source': (str, 'unknown'),
            'transition':    (str, 'unknown'),
            'element':  (str, 'unknown'),
            'charge':  (int, 0),
            'ion':  (str, 'unknown'),
            'symbol':   (str, 'unknown'),
        },
    }
    _forced_group = [_GROUP_NE, _GROUP_TE]
    _data_none = True

    _show_in_summary_core = ['shape', 'ref', 'group']
    _show_in_summary = 'all'

    _grouplines = _GROUP_LINES
    _groupne = _GROUP_NE
    _groupte = _GROUP_TE

    _units_lambda0 = _UNITS_LAMBDA0

    def add_line(
        self,
        key=None,
        lambda0=None,
        pec=None,
        source=None,
        transition=None,
        ion=None,
        symbol=None,
        **kwdargs,
    ):
        """ Add a spectral line by key and rest wavelength, optionally with

        """
        self.add_obj(
            which='lines',
            key=key,
            lambda0=lambda0,
            pec=pec,
            source=source,
            transition=transition,
            ion=ion,
            symbol=symbol,
            **kwdargs,
        )

    def add_pec(self, key=None, pec=None, ref=None):
        pass

    def remove_line(
        self,
        key=None,
        ion=None,
        source=None,
        lambda0=None,
    ):
        # Check inputs
        lc = [
            key is not None,
            ion is not None or source is not None or lambda0 is not None,
        ]
        raise NotImplementedError

    # -----------------
    # from openadas
    # ------------------

    @classmethod
    def _from_openadas(
        cls,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        online=None,
        update=None,
        create_custom=None,
        dsource0=None,
        dref0=None,
        ddata0=None,
        dlines0=None,
        grouplines=None,
    ):
        """
        Load lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/

        Provide wavelengths in m

        Example:
        --------
                >>> import tofu as tf
                >>> lines_mo = tf.data.SpectralLines.from_openadas(
                    element='Mo',
                    lambmin=3.94e-10,
                    lambmax=4e-10,
                )

        """

        # Preliminary import and checks
        from ..openadas2tofu import _requests
        from ..openadas2tofu import _read_files

        if online is None:
            online = _OPENADAS_ONLINE
        if grouplines is None:
            grouplines = cls._grouplines
        else:
            cls._grouplines = grouplines

        # Load from online if relevant
        if online is True:
            try:
                out = _requests.step01_search_online_by_wavelengthA(
                    lambmin=lambmin*1e10,
                    lambmax=lambmax*1e10,
                    element=element,
                    charge=charge,
                    verb=False,
                    returnas=np.ndarray,
                    resolveby='file',
                )
                lf = sorted(set([oo[0] for oo in out]))
                out = _requests.step02_download_all(
                    files=lf,
                    update=update,
                    create_custom=create_custom,
                    verb=False,
                )
            except Exception as err:
                msg = (
                    """
                    {}

                    For some reason data could not be downloaded from openadas
                        => see error message above
                        => maybe check your internet connection?
                    """.format(err)
                )
                raise Exception(msg)

        # Load for local files
        dne, dte, dpec, lion, dsource, dlines = _read_files.step03_read_all(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            pec_as_func=False,
            format_for_DataCollection=True,
            dsource0=dsource0,
            dref0=dref0,
            ddata0=ddata0,
            dlines0=dlines0,
            verb=False,
        )

        # # dgroup
        # dgroup = ['Te', 'ne']

        # dref - Te + ne
        dref = dte
        dref.update(dne)

        # ddata - pec
        ddata = dpec

        # dref_static
        dref_static = {
            'ion': {k0: {} for k0 in lion},
            'source': dsource,
        }

        # dobj (lines)
        dobj = {
            grouplines: dlines,
        }
        return ddata, dref, dref_static, dobj

    @classmethod
    def from_openadas(
        cls,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        online=None,
        update=None,
        create_custom=None,
        grouplines=None,
    ):
        """
        Load lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        ddata, dref, dref_static, dobj = cls._from_openadas(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            online=online,
            update=update,
            create_custom=create_custom,
            grouplines=grouplines,
        )
        return cls(ddata=ddata, dref=dref, dref_static=dref_static, dobj=dobj)

    def add_from_openadas(
        self,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        online=None,
        update=None,
        create_custom=None,
    ):
        """
        Load and add lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        ddata, dref, dref_static, dobj = self._from_openadas(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            online=online,
            update=update,
            create_custom=create_custom,
            dsource0=self._dref_static.get('source'),
            dref0=self._dref,
            ddata0=self._ddata,
            dlines0=self._dobj.get('lines'),
        )
        self.update(ddata=ddata, dref=dref, dref_static=dref_static, dobj=dobj)

    # -----------------
    # from nist
    # ------------------

    @classmethod
    def _from_nist(
        cls,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        ion=None,
        wav_observed=None,
        wav_calculated=None,
        transitions_allowed=None,
        transitions_forbidden=None,
        cache_from=None,
        cache_info=None,
        verb=None,
        create_custom=None,
        dsource0=None,
        dlines0=None,
        grouplines=None,
    ):
        """
        Load lines from nist, either:
            - cache_from = False:  directly from the website
            - cache_from = True: from pre-downloaded files in ~/.tofu/nist/

        Provide wavelengths in m

        Example:
        --------
                >>> import tofu as tf
                >>> lines_mo = tf.data.SpectralLines.from_nist(
                    element='Mo',
                    lambmin=3.94e-10,
                    lambmax=4e-10,
                )

        """

        # Preliminary import and checks
        from ..nist2tofu import _requests

        if grouplines is None:
            grouplines = cls._grouplines
        else:
            cls._grouplines = grouplines
        if verb is None:
            verb = False
        if cache_info is None:
            cache_info = False

        # Load from online if relevant
        dlines, dsources = _requests.step01_search_online_by_wavelengthA(
            element=element,
            charge=charge,
            ion=ion,
            lambmin=lambmin*1e10,
            lambmax=lambmax*1e10,
            wav_observed=wav_observed,
            wav_calculated=wav_calculated,
            transitions_allowed=transitions_allowed,
            transitions_forbidden=transitions_forbidden,
            info_ref=True,
            info_conf=True,
            info_term=True,
            info_J=True,
            info_g=True,
            cache_from=cache_from,
            cache_info=cache_info,
            return_dout=True,
            return_dsources=True,
            verb=verb,
            create_custom=create_custom,
            format_for_DataCollection=True,
            dsource0=dsource0,
            dlines0=dlines0,
        )

        # dref_static
        lion = sorted(set([dlines[k0]['ion'] for k0 in dlines.keys()]))

        dref_static = {
            'ion': {k0: {} for k0 in lion},
            'source': dsources,
        }

        # dobj (lines)
        dobj = {
            grouplines: dlines,
        }
        return dref_static, dobj

    @classmethod
    def from_nist(
        cls,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        ion=None,
        wav_observed=None,
        wav_calculated=None,
        transitions_allowed=None,
        transitions_forbidden=None,
        cache_from=None,
        cache_info=None,
        verb=None,
        create_custom=None,
        grouplines=None,
    ):
        """
        Load lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        dref_static, dobj = cls._from_nist(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            ion=ion,
            wav_observed=wav_observed,
            wav_calculated=wav_calculated,
            transitions_allowed=transitions_allowed,
            transitions_forbidden=transitions_forbidden,
            cache_from=cache_from,
            cache_info=cache_info,
            verb=verb,
            create_custom=create_custom,
            grouplines=grouplines,
        )
        return cls(dref_static=dref_static, dobj=dobj)

    def add_from_nist(
        self,
        lambmin=None,
        lambmax=None,
        element=None,
        charge=None,
        ion=None,
        wav_observed=None,
        wav_calculated=None,
        transitions_allowed=None,
        transitions_forbidden=None,
        cache_from=None,
        cache_info=None,
        verb=None,
        create_custom=None,
        grouplines=None,
    ):
        """
        Load and add lines and pec from openadas, either:
            - online = True:  directly from the website
            - online = False: from pre-downloaded files in ~/.tofu/openadas/
        """
        dref_static, dobj = self._from_nist(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element,
            charge=charge,
            ion=ion,
            wav_observed=wav_observed,
            wav_calculated=wav_calculated,
            transitions_allowed=transitions_allowed,
            transitions_forbidden=transitions_forbidden,
            cache_from=cache_from,
            cache_info=cache_info,
            verb=verb,
            create_custom=create_custom,
            dsource0=self._dref_static.get('source'),
            dlines0=self._dobj.get('lines'),
            grouplines=grouplines,
        )
        self.update(dref_static=dref_static, dobj=dobj)

    # -----------------
    # from file (.py)
    # ------------------

    @staticmethod
    def _check_extract_dict_from_mod(mod, k0):
        lk1 = [
            k0, k0.upper(),
            '_'+k0, '_'+k0.upper(),
            '_d'+k0, '_D'+k0.upper(),
            'd'+k0, 'D'+k0.upper(),
            k0+'s', k0.upper()+'S'
            '_d'+k0+'s', '_D'+k0.upper()+'S',
            'd'+k0+'s', 'D'+k0.upper()+'S',
        ]
        lk1 = [k1 for k1 in lk1 if hasattr(mod, k1)]
        if len(lk1) > 1:
            msg = "Ambiguous attributes: {}".format(lk1)
            raise Exception(msg)
        elif len(lk1) == 0:
            return

        if hasattr(mod, lk1[0]):
            return getattr(mod, lk1[0])
        else:
            return

    @classmethod
    def from_module(cls, pfe=None):

        # Check input
        c0 = (
            os.path.isfile(pfe)
            and pfe[-3:] == '.py'
        )
        if not c0:
            msg = (
                "\nProvided Path-File-Extension (pfe) not valid!\n"
                + "\t- expected: absolute path to python module\n"
                + "\t- provided: {}".format(pfe)
            )
            raise Exception(msg)
        pfe = os.path.abspath(pfe)

        # Load module
        path, fid = os.path.split(pfe)
        import importlib.util
        spec = importlib.util.spec_from_file_location(fid[:-3], pfe)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

        # extract dref_static
        dref_static = {}
        for k0 in ['source', 'transition', 'ion', 'element']:
            dd = cls._check_extract_dict_from_mod(mod, k0)
            if dd is not None:
                dref_static[k0] = dd

        # add ion
        if 'ion' not in dref_static.keys():
            lions = np.array([
                    v0['ion'] for k0, v0 in mod.dlines.items()
                    if 'ion' in v0.keys()
            ]).ravel()
            if len(lions) > 0:
                dref_static['ion'] = {
                    k0: {'ion': k0} for k0 in lions
                }
            else:
                lIONS = np.array([
                        v0['ION'] for k0, v0 in mod.dlines.items()
                        if 'ION' in v0.keys()
                ]).ravel()
                if len(lIONS) > 0:
                    dref_static['ION'] = {
                        k0: {'ION': k0} for k0 in lIONS
                    }

        # extract lines
        dobj = {
            'lines': mod.dlines
        }

        # Create collection
        out = cls(dref_static=dref_static, dobj=dobj)

        # Replace ION by ion if relevant
        c0 = (
            'ion' in out.dref_static.keys()
            and 'ion' not in out.get_lparam(which='lines')
            and 'ION' in out.get_lparam(which='lines')
        )
        if c0:
            for k0, v0 in out._dobj['lines'].items():
                ion = [
                    k1 for k1, v1 in out._dref_static['ion'].items()
                    if out._dobj['lines'][k0]['ION'] == v1['ION']
                ][0]
                out._dobj['lines'][k0]['ion'] = ion
                del out._dobj['lines'][k0]['ION']
        return out

    # -----------------
    # summary
    # ------------------

    # -----------------
    # conversion wavelength - energy - frequency
    # ------------------

    def convert_lines(self, units=None, key=None, ind=None, returnas=None):
        """ Convert wavelength (m) to other units or other quantities

        Avalaible units:
            wavelength: km, m, mm, um, nm, pm, A
            energy:     J, eV, keV, MeV, GeV
            frequency:  Hz, kHz, MHz, GHz, THz

        Return the result as a np.ndarray (returnas = 'data')

        Can also just return the conversion coef if returnas='coef'
        In that case, a bool is also returned indicating whether the result is
        the proportional to the inverse of lambda0::
            - False: data = coef * lambda0
            - True: data = coef / lambda0
        """
        if units is None:
            units = self._units_lambda0
        if returnas is None:
            returnas = dict

        lok = [dict, np.ndarray, 'data', 'coef']
        if returnas not in lok:
            msg = (
                "Arg returnas must be in:\n"
                + "\t- {}\n".format(lok)
                + "\t- provided: {}".format(returnas)
            )
            raise Exception(msg)
        if returnas in [dict, np.ndarray, 'data']:
            returnas2 = 'data'
        else:
            returnas2 = 'coef'

        key = self._ind_tofrom_key(
            which=self._grouplines, key=key, ind=ind, returnas=str,
        )
        lamb_in = self.get_param(
            'lambda0', key=key, returnas=np.ndarray,
        )['lambda0']
        out = self.convert_spectral(
            data=lamb_in, units_in='m', units_out=units, returnas=returnas2,
        )
        if returnas is dict:
            out = {k0: out[ii] for ii, k0 in enumerate(key)}
        return out

    # -----------------
    # PEC interpolation
    # ------------------

    def calc_pec(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        deg=None,
        grid=None,
    ):
        """ Compute the pec (<sigma v>) by interpolation for chosen lines

        Assumes Maxwellian electron distribution

        Provide ne and Te and 1d np.ndarrays

        if grid=False:
            - ne is a (n,) 1d array
            - Te is a (n,) 1d array
          => the result is a dict of (n,) 1d array

        if grid=True:
            - ne is a (n1,) 1d array
            - Te is a (n2,) 1d array
          => the result is a dict of (n1, n2) 2d array
        """

        # Check keys
        key = self._ind_tofrom_key(
            which=self._grouplines, key=key, ind=ind, returnas=str,
        )
        dlines = self._dobj[self._grouplines]

        if deg is None:
            deg = 2

        # Check data conformity
        lg = (self._groupne, self._groupte)
        lc = [
            k0 for k0 in key
            if (
                dlines[k0].get('pec') is None
                or self._ddata[dlines[k0]['pec']]['group'] != lg
            )
        ]
        if len(lc) > 0:
            msg = (
                "The following lines have non-conform pec data:\n"
                + "\t- {}\n\n".format(lc)
                + "  => pec data should be tabulated vs (ne, Te)"
            )
            warnings.warn(msg)
            key = [kk for kk in key if kk not in lc]

        # Check ne, Te
        ltype = [int, float, np.integer, np.floating]
        dnTe = {'ne': ne, 'Te': Te}
        for k0, v0 in dnTe.items():
            if type(v0) in ltype:
                dnTe[k0] = np.r_[v0]
            if isinstance(dnTe[k0], list) or isinstance(dnTe[k0], tuple):
                dnTe[k0] = np.array([dnTe[k0]])
            if not (isinstance(dnTe[k0], np.ndarray) and dnTe[k0].ndim == 1):
                msg = (
                    "Arg {} should be a 1d np.ndarray!".format(k0)
                )
                raise Exception(msg)

        # Interpolate
        dout = {}
        derr = {}
        for k0 in key:
            try:
                ne0 = [
                    kk for kk in self._ddata[dlines[k0]['pec']]['ref']
                    if self._ddata[kk]['group'] == (self._groupne,)
                ][0]
                ne0 = self._ddata[ne0]['data']
                Te0 = [
                    kk for kk in self._ddata[dlines[k0]['pec']]['ref']
                    if self._ddata[kk]['group'] == (self._groupte,)
                ][0]
                Te0 = self._ddata[Te0]['data']

                dout[k0] = _comp_spectrallines._interp_pec(
                    ne0=ne0,
                    Te0=Te0,
                    pec0=self._ddata[dlines[k0]['pec']]['data'],
                    ne=dnTe['ne'],
                    Te=dnTe['Te'],
                    deg=deg,
                    grid=grid,
                )
            except Exception as err:
                derr[k0] = str(err)

        if len(derr) > 0:
            msg = (
                "The pec could not be interpolated for the following lines:\n"
                + "\n".join([
                    '\t- {} : {}'.format(k0, v0) for k0, v0 in derr.items()
                ])
            )
            raise Exception(msg)

        return dout

    def calc_intensity(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        concentration=None,
        deg=None,
        grid=None,
    ):
        """ Compute the lines intensities by pec interpolation for chosen lines

        Assumes Maxwellian electron distribution

        Provide ne and Te and 1d np.ndarrays

        Provide concentration as:
            - a np.ndarray (same concentration assumed for all lines)
            - a dict of {key: np.ndarray}

        if grid=False:
            - ne is a (n,) 1d array
            - Te is a (n,) 1d array
            - concentration is a (dict of) (n,) 1d array(s)
          => the result is a dict of (n1, n2) 2d array

        if grid=True:
            - ne is a (n1,) 1d array
            - Te is a (n2,) 1d array
            - concentration is a (dict of) (n1, n2) 2d array(s)
          => the result is a dict of (n1, n2) 2d array


        """

        # check inputs
        if grid is None:
            grid = ne.size != Te.size

        # Check keys
        key = self._ind_tofrom_key(
            which=self._grouplines, key=key, ind=ind, returnas=str,
        )

        if isinstance(concentration, np.ndarray):
            concentration = {k0: concentration for k0 in key}
        c0 = (
            isinstance(concentration, dict)
            and all([
                k0 in key
                and isinstance(cc, np.ndarray)
                and (
                    (grid is False and cc.shape == ne.shape == Te.shape)
                    or
                    (grid is True and cc.shape == (ne.size, Te.size))
                )
                and np.all((cc > 0.) & (cc <= 1.))
                for k0, cc in concentration.items()
            ])
        )
        if not c0:
            shape = ne.shape if grid is False else (ne.size, Te.size)
            msg = (
                "Arg concentration is non-conform:\n"
                + "\t- Expected: dict of {} arrays in [0, 1]\n".format(shape)
                + "\t- Provided: {}".format(concentration)
            )
            raise Exception(msg)

        # interpolate pec
        dpec = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne,
            Te=Te,
            grid=grid,
            deg=deg,
        )

        # ne for broadcasting
        if grid is True:
            neb = ne[:, None]
        else:
            neb = ne

        # Derive intensity
        dint = {
            k0: v0*neb**2*concentration[k0] for k0, v0 in dpec.items()
        }
        return dint

    # -----------------
    # plotting
    # ------------------

    def plot(
        self,
        key=None,
        ind=None,
        ax=None,
        sortby=None,
        param_txt=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dcolor=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        wintit=None,
        tit=None,
    ):
        """ plot rest wavelengths as vertical lines """
        if param_txt is None:
            param_txt = 'symbol'

        return super()._plot_axvlines(
            which='lines',
            key=key,
            param_x='lambda0',
            param_txt=param_txt,
            sortby=sortby,
            sortby_def='ion',
            sortby_lok=['ion', 'ION', 'source'],
            ax=ax, ymin=ymin, ymax=ymax,
            ls=ls, lw=lw, fontsize=fontsize,
            side=side, dcolor=dcolor, fraction=fraction,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit,
        )

    def plot_pec_single(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        concentration=None,
        deg=None,
        grid=None,
        ax=None,
        sortby=None,
        param_txt=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dcolor=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        wintit=None,
        tit=None,
    ):

        # Check input
        if param_txt is None:
            param_txt = 'symbol'

        # Check ne, Te
        ltypes = [int, float, np.integer, np.floating]
        dnTe = {'ne': ne, 'Te': Te}
        single = all([
            type(v0) in ltypes or len(v0) == 1 for v0 in dnTe.values()
        ])
        if not single:
            msg = ("Arg ne and Te must be floats!")
            raise Exception(msg)

        # Get dpec
        dpec = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne,
            Te=Te,
            deg=deg,
            grid=grid,
        )
        key = list(dpec.keys())

        ne = float(ne)
        Te = float(Te)
        tit = (
            r'$n_e$' + '= {} '.format(ne) + r'$/m^3$'
            + r' -  $T_e$ = ' + '{} keV'.format(Te/1000.)
        )

        pmax = np.max([np.log10(v0) for v0 in dpec.values()])
        pmin = np.min([np.log10(v0) for v0 in dpec.values()])
        dsize = {
            k0: (np.log10(v0)-pmin)/(pmax-pmin)*19 + 1
            for k0, v0 in dpec.items()
        }

        sortby_lok = ['ion', 'ION', 'source']
        lk0 = [k0 for k0 in sortby_lok if k0 in self._dref_static.keys()]
        if len(lk0) > 0:
            sortby_def = lk0[0]
        else:
            sortby_def = None

        return super()._plot_axvlines(
            which='lines',
            key=key,
            param_x='lambda0',
            param_txt=param_txt,
            sortby=sortby,
            sortby_def=sortby_def,
            sortby_lok=sortby_lok,
            dsize=dsize,
            ax=ax, ymin=ymin, ymax=ymax,
            ls=ls, lw=lw, fontsize=fontsize,
            side=side, dcolor=dcolor, fraction=fraction,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit,
        )

    def plot_pec(
        self,
        key=None,
        ind=None,
        ne=None,
        Te=None,
        norder=None,
        ne_scale=None,
        Te_scale=None,
        param_txt=None,
        param_color=None,
        deg=None,
        dax=None,
        proj=None,
        ymin=None,
        ymax=None,
        ls=None,
        lw=None,
        fontsize=None,
        side=None,
        dcolor=None,
        fraction=None,
        figsize=None,
        dmargin=None,
        dtit=None,
        tit=None,
        wintit=None,
    ):

        # Check input
        if param_txt is None:
            param_txt = 'symbol'
        if param_color is None:
            param_color = 'ion'
        if norder is None:
            norder = 0

        if ne_scale is None:
            ne_scale = 'log'
        if Te_scale is None:
            Te_scale = 'linear'

        # Check ne, Te
        ltypes = [int, float, np.integer, np.floating]
        dnTe = {
            'ne': type(ne) in ltypes or len(ne) == 1,
            'Te': type(Te) in ltypes or len(Te) == 1,
        }
        if all([v0 for v0 in dnTe.values()]):
            msg = (
                "For a single point in (ne, Te) space, use plot_pec_singe()"
            )
            raise Exception(msg)
        elif dnTe['ne']:
            ne = np.r_[ne].ravel()
            ne = np.full((Te.size), ne[0])
        elif dnTe['Te']:
            Te = np.r_[Te].ravel()
            Te = np.full((ne.size), Te[0])

        if len(ne) != len(Te):
            msg = (
                "Please provide ne and Te as vectors of same size!"
            )
            raise Exception(msg)

        # Get dpec
        dpec = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne,
            Te=Te,
            deg=deg,
            grid=False,
        )
        damp = {k0: {'data': v0} for k0, v0 in dpec.items()}

        # Create grid
        ne_grid = _DataCollection_comp._get_grid1d(
            ne, scale=ne_scale, npts=ne.size*2, nptsmin=3,
        )
        Te_grid = _DataCollection_comp._get_grid1d(
            Te, scale=Te_scale, npts=Te.size*2, nptsmin=3,
        )

        dpec_grid = self.calc_pec(
            key=key,
            ind=ind,
            ne=ne_grid,
            Te=Te_grid,
            deg=deg,
            grid=True,
        )

        # Get dcolor
        lcol = plt.rcParams['axes.prop_cycle'].by_key()['color']
        dcolor = {}
        if param_color != 'key':
            lion = [self._dobj['lines'][k0][param_color] for k0 in dpec.keys()]
            for ii, k0 in enumerate(set(lion)):
                dcolor[k0] = mcolors.to_rgb(lcol[ii % len(lcol)])
                lk1 = [
                    k2 for k2 in dpec.keys()
                    if self._dobj['lines'][k2][param_color] == k0
                ]
                for k1 in lk1:
                    damp[k1]['color'] = k0
        else:
            for ii, k0 in enumerate(dpec.keys()):
                dcolor[k0] = mcolors.to_rgb(lcol[ii % len(lcoil)])
                damp[k0]['color'] = k0

        # Create image
        im_data = np.full((ne_grid.size, Te_grid.size), np.nan)
        im = np.full((ne_grid.size, Te_grid.size, 4), np.nan)
        dom_val = np.concatenate(
            [v0[None, :, :] for v0 in dpec_grid.values()],
            axis=0,
        )

        if norder == 0:
            im_ind = np.nanargmax(dom_val, axis=0)
        else:
            im_ind = np.argsort(dom_val, axis=0)[-norder, :, :]

        for ii in np.unique(im_ind):
            ind = im_ind == ii
            im_data[ind] = dom_val[ii, ind]

        pmin = np.nanmin(np.log10(im_data))
        pmax = np.nanmax(np.log10(im_data))

        for ii, k0 in enumerate(dpec_grid.keys()):
            if ii in np.unique(im_ind):
                ind = im_ind == ii
                im[ind, :-1] = dcolor[damp[k0]['color']]
                im[ind, -1] = (
                    (np.log10(im_data[ind])-pmin)/(pmax-pmin)*0.9 + 0.1
                )
        extent = (ne_grid.min(), ne_grid.max(), Te_grid.min(), Te_grid.max())

        if tit is None:
            tit = 'spectral lines PEC interpolations'
        if dtit is None:
            dtit = {'map': 'norder = {}'.format(norder)}

        return _DataCollection_plot.plot_dominance_map(
            din=self._dobj['lines'], im=im, extent=extent,
            xval=ne, yval=Te, damp=damp,
            x_scale=ne_scale, y_scale=Te_scale, amp_scale='log',
            param_txt='symbol',
            dcolor=dcolor,
            dax=dax, proj=proj,
            figsize=figsize, dmargin=dmargin,
            wintit=wintit, tit=tit, dtit=dtit,
        )


#############################################
#############################################
#       Time traces
#############################################


# TBC
class TimeTraces(DataCollection):
    """ A generic class for handling multiple time traces """

    _forced_group = 'time'
    _dallowed_params = {
        'time': {
            'origin': (str, 'unknown'),
            'dim':    (str, 'time'),
            'quant':  (str, 't'),
            'name':   (str, 't'),
            'units':  (str, 's')},
    }
    _plot_vignettes = False

    def fit(self, ind=None, key=None,
            Type='staircase', func=None,
            plot=True, fs=None, ax=None, draw=True, **kwdargs):
        """  Fit the times traces with a model

        Typically try to fit plateaux and ramps i.e.: Type = 'staircase')
        Return a dictionary of the fitted parameters, ordered by data key

        """

        dout = self._fit_one_dim(ind=ind, key=key, group=self._forced_group,
                                 Type=Type, func=func, **kwdargs)
        if plot:
            kh = _DataCollection_plot.plot_fit_1d(self, dout)
        return dout

    def add_plateaux(self, verb=False):
        dout = self.fit(
            ind=ind, key=key, group=group, Type='staircase',
        )

        # Make Pandas Dataframe attribute
        self.plateaux = None
        if verb:
            msg = ""

    def plot(self, **kwdargs):
        return self._plot_timetraces(**kwdargs)

    def plot_incremental(self, key=None, ind=None,
                         plateaux=True, connect=True):
        return

    def plot_plateau_validate(self, key=None, ind=None):
        return




"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import sys
import warnings
import itertools as itt

# Standard
import numpy as np
import scipy.constants as scpct
import matplotlib.pyplot as plt

# tofu-specific
from tofu import __version__
import tofu.spectro as tfs

_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.spectro.fit12d'


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    print("")   # this is to get a newline after the dots
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if not lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following previous test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    # print("setup_module before anything in this file")


def teardown_module(module):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    pass


# def my_setup_function():
#    print ("my_setup_function")

# def my_teardown_function():
#    print ("my_teardown_function")

# @with_setup(my_setup_function, my_teardown_function)
# def test_numbers_3_4():
#    print 'test_numbers_3_4  <============================ actual test code'
#    assert multiply(3,4) == 12

# @with_setup(my_setup_function, my_teardown_function)
# def test_strings_a_3():
#    print 'test_strings_a_3  <============================ actual test code'
#    assert multiply('a',3) == 'aaa'


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################


class Test01_DataCollection(object):

    @classmethod
    def setup_class(cls):

        nlamb = 100
        lamb = np.linspace(3.94, 4, nlamb)*1e-10
        var = np.linspace(-25, 25, 51)

        def bck(lamb=None, offset=None, slope=None):
            return offset + (lamb-lamb.min())*slope/(lamb.max()-lamb.min())

        def gauss(lamb=None, lamb0=None, sigma=None, delta=None, amp=None):
            return amp * np.exp(-(lamb-lamb0-delta)**2/(2*sigma**2))

        def noise(lamb=None, amp=None, freq=None, phase=None):
            return amp*np.sin(
                lamb*freq*2.*np.pi/(lamb.max()-lamb.min()) + phase
            )

        dlines = {
            'a': {
                'lambda0': 3.95e-10,
                'delta': 0.001e-10,
                'sigma': 0.002e-10,
                'amp': 1.,
                'noise': 0.01,
                'group': 0,
            },
            'b': {
                'lambda0': 3.97e-10,
                'delta': -0.001e-10,
                'sigma': 0.0015e-10,
                'amp': 0.5,
                'noise': 0.001,
                'group': 1,
            },
            'c': {
                'lambda0': 3.975e-10,
                'delta': 0.001e-10,
                'sigma': 0.001e-10,
                'amp': 0.6,
                'noise': 0.005,
                'group': 0,
            },
            'd': {
                'lambda0': 3.99e-10,
                'delta': 0.002e-10,
                'sigma': 0.002e-10,
                'amp': 0.8,
                'noise': 0.01,
                'group': 1,
            },
        }

        spect2d = bck(
            lamb=lamb[None, :],
            offset=0.1*np.exp(-(var[:, None]-25)**2/10**2),
            slope=0.001,
        )
        spect2d += noise(lamb=lamb[None, :], amp=0.01, freq=10, phase=0.)

        for ii, k0 in enumerate(dlines.keys()):
            spect2d += gauss(
                lamb=lamb[None, :],
                amp=dlines[k0]['amp'] * np.exp(-var[:, None]**2/20**2),
                lamb0=dlines[k0]['lambda0'],
                sigma=dlines[k0]['sigma']*(
                    1 + 2*(ii/len(dlines))*np.cos(var[:, None]*2*np.pi/50)
                ),
                delta=dlines[k0]['delta']*(
                    1 + 2*(ii/len(dlines))*np.sin(
                        var[:, None]*2*np.pi*(len(dlines)-ii)/50
                    )
                ),
            )
            spect2d += noise(
                lamb=lamb[None, :],
                amp=dlines[k0]['noise'] * np.exp(-var[:, None]**2/10**2),
                freq=10*(len(dlines)-ii),
                phase=ii,
            )

        mask = np.repeat((np.abs(var-15) < 3)[:, None], nlamb, axis=1)

        # Plot spect 2d
        # fig = plt.figure(figsize=(12, 10));
        # ax0 = fig.add_axes([0.05, 0.1, 0.4, 0.8])
        # ax0.set_xlabel(r'$\lambda$ (m)')
        # ax0.set_ylabel(r'$y$ (m)')
        # extent = (lamb.min(), lamb.max(), var.min(), var.max())
        # ax0.imshow(spect2d, extent=extent, aspect='auto', origin='lower');
        # sp2bis = np.copy(spect2d)
        # sp2bis[mask] = np.nan
        # ax1 = fig.add_axes([0.55, 0.1, 0.4, 0.8])
        # ax1.set_xlabel(r'$\lambda$ (m)')
        # ax1.set_ylabel(r'$y$ (m)')
        # ax1.imshow(sp2bis, extent=extent, aspect='auto', origin='lower');
        # plt.ion()
        # plt.show();
        # import pdb; pdb.set_trace()     # DB

        cls.lamb = lamb
        cls.var = var
        cls.dlines = dlines
        cls.spect2d = spect2d
        cls.mask = mask
        cls.ldinput1d = []
        cls.ldfit1d = []
        cls.ldex1d = []
        cls.ldinput2d = []
        cls.ldfit2d = []
        cls.ldex2d = []

    @classmethod
    def setup(self):
        pass

    def teardown(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_fit1d_dinput(self):
        defconst = {
            'amp': False,
            'width': False,
            'shift': False,
            'double': False,
            'symmetry': False,
        }

        # Define constraint dict
        ldconst = [
            {
                'amp': {'a1': ['a', 'd']},
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'b': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's3', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': True,
                'symmetry': True,
            },
            {
                'amp': False,
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's2', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': False,
                'symmetry': False,
            },
            {
                'amp': False,
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's2', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': False,
                'symmetry': False,
            },
        ]

        ldx0 = [
            None,
            {
                # 'amp': {''},
                'width': 1.,
                'shift': {
                    's1': 0.,
                    's2': 1.,
                },
                'dratio': 0,
                'dshift': 0,
            }
        ]

        ldomain = [
            None,
            {
                'lamb': [
                    [3.94e-10, 3.952e-10],
                    (3.95e-10, 3.956e-10),
                    [3.96e-10, 4e-10],
                ],
            },
        ]

        ldata = [
            self.spect2d[5, :],
            self.spect2d[:5, :],
        ]
        lpos = [False, True]

        lfocus = [None, 'a', [3.94e-10, 3.96e-10]]

        ldconstants = [None, {'shift': {'s1': 0}}]

        combin = [ldconst, ldx0, ldomain, ldata, lpos, lfocus, ldconstants]
        for comb in itt.product(*combin):
            dinput = tfs.fit1d_dinput(
                dlines=self.dlines,
                dconstraints=comb[0],
                dconstants=comb[6],
                dprepare=None,
                data=np.copy(comb[3]),
                lamb=self.lamb,
                mask=None,
                domain=comb[2],
                pos=comb[4],
                subset=None,
                same_spectrum=None,
                nspect=None,
                same_spectrum_dlamb=None,
                focus=comb[5],
                valid_fraction=0.28,     # fraction of pixels ok per time step
                valid_nsigma=0,         # S/N ratio for each pixel
                focus_half_width=None,
                valid_return_fract=None,
                dscales=None,
                dx0=comb[1],
                dbounds=None,
                defconst=defconst,
            )
            self.ldinput1d.append(dinput)

    def test02_funccostjac_1d(self):
        func = tfs._fit12d_funccostjac.multigausfit1d_from_dlines_funccostjac
        for ii, dd in enumerate(self.ldinput1d):
            func_detail, func_cost, func_jac = func(
                lamb=dd['dprepare']['lamb'], dinput=dd,
                dind=dd['dind'], jac='dense',
            )

            # Get x0
            x0 = tfs._fit12d._dict2vector_dscalesx0bounds(
                dd=dd['dx0'], dd_name='dx0', dinput=dd,
            )
            scales = tfs._fit12d._dict2vector_dscalesx0bounds(
                dd=dd['dscales'], dd_name='dscales', dinput=dd,
            )

            y0 = func_detail(x0[0, :], scales=scales[0, :])
            y1 = func_cost(
                x0[0, :],
                scales=scales[0, :],
                data=dd['dprepare']['data'][0, :],
            )

            # check consistency between func_detail and func_cost
            assert np.allclose(
                np.sum(y0, axis=1) - dd['dprepare']['data'][0, :],
                y1,
                equal_nan=True,
            )

    def test03_fit1d(self):
        lchain = [False, True]
        for ii, dd in enumerate(itt.product(self.ldinput1d, lchain)):
            dfit1d = tfs.fit1d(
                dinput=dd[0],
                method=None,
                Ti=None,
                chain=dd[1],
                jac='dense',
                verbose=None,
                plot=False,
            )
            assert np.sum(dfit1d['validity'] < 0) == 0
            self.ldfit1d.append(dfit1d)

    def test04_fit1d_dextract(self):
        for ii, dd in enumerate(self.ldfit1d):
            dex = tfs.fit1d_extract(
                dfit1d=dd,
                ratio=('a', 'c'),
                pts_lamb_detail=True,
            )
            self.ldex1d.append(dex)

    def test05_fit1d_plot(self):
        lwar = []
        for ii, dd in enumerate(self.ldex1d):
            try:
                # For a yet unknown reason, this particular test crahses on
                # Windows only due to figure creation at
                # tfs._plot.plot_fit1d(): line 337
                # already investigated: reducing figure size and early closing
                # No more ideas...
                # This link suggests it may have something to do with 
                # inches => pixels conversion of figure size...
                # https://github.com/matplotlib/matplotlib/issues/14225
                if 'win' not in sys.platform.lower():
                    dax = tfs._plot.plot_fit1d(
                        dfit1d=self.ldfit1d[ii],
                        dextract=dd,
                        annotate=self.ldfit1d[ii]['dinput']['keys'][0],
                        fs=(4, 4),
                    )
            except Exception as err:
                lwar.append((ii, str(err)))
            finally:
                plt.close('all')

        if len(lwar) > 0:
            msg = (
                "\nThe ({}/{}) following fit1d plots failed:\n".format(
                    len(lwar), len(self.ldex1d),
                )
                + "\n".join(["\t- {}: {}".format(ww[0], ww[1]) for ww in lwar])
            )
            warnings.warn(msg)

    def test06_fit2d_dinput(self):
        defconst = {
            'amp': False,
            'width': False,
            'shift': False,
            'double': False,
            'symmetry': False,
        }

        # Define constraint dict
        ldconst = [
            {
                'amp': {'a1': ['a', 'd']},
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'b': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's3', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': True,
                'symmetry': True,
            },
            {
                'amp': False,
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's2', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': False,
                'symmetry': False,
            },
            {
                'amp': False,
                'width': 'group',
                'shift': {
                    'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                    'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                    'd': {'key': 's2', 'coef': 1., 'offset': 0.001e-10},
                },
                'double': False,
                'symmetry': False,
            },
        ]

        ldx0 = [
            None,
            {
                # 'amp': {''},
                'width': 1.,
                'shift': {
                    's1': 0.,
                    's2': 1.,
                },
                'dratio': 0,
                'dshift': 0,
            }
        ]

        ldomain = [
            None,
            {
                'lamb': [
                    [3.94e-10, 3.952e-10],
                    (3.95e-10, 3.956e-10),
                    [3.96e-10, 4e-10],
                ],
            },
        ]

        lpos = [False, True]
        lmask = [None, self.mask]

        lfocus = [None, 'a', [3.94e-10, 3.96e-10]]

        ldconstants = [None, {'shift': {'s1': 0}}]

        combin = [ldconst, ldx0, ldomain, lmask, lpos, lfocus, ldconstants]
        for comb in itt.product(*combin):
            dinput = tfs.fit2d_dinput(
                dlines=self.dlines,
                dconstraints=comb[0],
                dconstants=comb[6],
                dprepare=None,
                data=np.copy(self.spect2d),
                lamb=self.lamb,
                phi=self.var,
                mask=comb[3],
                nbsplines=5,
                domain=comb[2],
                pos=comb[4],
                subset=None,
                focus=comb[5],
                valid_fraction=0.28,     # fraction of pixels ok per time step
                valid_nsigma=0,         # S/N ratio for each pixel
                focus_half_width=None,
                valid_return_fract=None,
                dscales=None,
                dx0=comb[1],
                dbounds=None,
                defconst=defconst,
            )
            self.ldinput2d.append(dinput)

    """
    def test07_funccostjac_2d(self):
        func = tfs._fit12d_funccostjac.multigausfit2d_from_dlines_funccostjac
        for ii, dd in enumerate(self.ldinput2d):
            func_detail, func_cost, func_jac = func(
                lamb=dd['dprepare']['lamb'],
                phi=dd['dprepare']['phi'],
                dinput=dd,
                dind=dd['dind'], jac='dense',
            )

            # Get x0
            x0 = tfs._fit12d._dict2vector_dscalesx0bounds(
                dd=dd['dx0'], dd_name='dx0', dinput=dd,
            )
            scales = tfs._fit12d._dict2vector_dscalesx0bounds(
                dd=dd['dscales'], dd_name='dscales', dinput=dd,
            )

            y0 = func_detail(x0[0, :], scales=scales[0, :])
            y1 = func_cost(
                x0[0, :],
                scales=scales[0, :],
                data=dd['dprepare']['data'][0, :],
            )

            # check consistency between func_detail and func_cost
            assert np.allclose(
                np.sum(y0, axis=1) - dd['dprepare']['data'][0, :],
                y1,
                equal_nan=True,
            )
    """



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.interpolate as scpinterp


# specific
from . import _generic_check
from . import _mesh_bsplines_operators


# #############################################################################
# #############################################################################
#                       BivariateSplineRect - scipy subclass
# #############################################################################


class BivariateSplineRect(scpinterp.BivariateSpline):
    """ Subclass for tofu

    Defined from knots (unique) and deg
    coefs set to 1 by default

    Used self.set_coefs() to update
    """

    def __init__(self, knotsR=None, knotsZ=None, deg=None, shapebs=None):

        assert np.allclose(np.unique(knotsR), knotsR)
        assert np.allclose(np.unique(knotsZ), knotsZ)
        assert deg in [0, 1, 2, 3]

        # get knots pr bs
        self._get_knots_per_bs_for_basis_elements(
            knotsR=knotsR,
            knotsZ=knotsZ,
            deg=deg,
        )

        # full knots with multiplicity
        knotsR, nbsR = _get_bs2d_func_knots(
            knotsR, deg=deg, returnas='data', return_unique=True,
        )
        knotsZ, nbsZ = _get_bs2d_func_knots(
            knotsZ, deg=deg, returnas='data', return_unique=True,
        )

        coefs = np.ones((nbsR*nbsZ,), dtype=float)

        self.__nbs = (nbsR, nbsZ)
        self.tck = [knotsR, knotsZ, coefs]
        self.degrees = [deg, deg]

        # shapebs
        self.shapebs = shapebs

    def _get_knots_per_bs_for_basis_elements(
        self,
        knotsR=None,
        knotsZ=None,
        deg=None,
    ):

        # added for details
        knots_per_bs_x = _get_bs2d_func_knots(knotsR, deg=deg, returnas='data')
        knots_per_bs_y = _get_bs2d_func_knots(knotsZ, deg=deg, returnas='data')

        self.knots_per_bs_x = knots_per_bs_x
        self.knots_per_bs_y = knots_per_bs_y

        if deg == 0:
            pass
        else:
            knots_per_bs_x = np.concatenate(
                (
                    np.tile(knots_per_bs_x[0, :] - 1, (deg, 1)),
                    knots_per_bs_x,
                    np.tile(knots_per_bs_x[-1, :] + 1, (deg, 1)),
                ),
                axis=0,
            )
            knots_per_bs_y = np.concatenate(
                (
                    np.tile(knots_per_bs_y[0, :] - 1, (deg, 1)),
                    knots_per_bs_y,
                    np.tile(knots_per_bs_y[-1, :] + 1, (deg, 1)),
                ),
                axis=0,
            )

        self.knots_per_bs_x_pad = np.asfortranarray(knots_per_bs_x)
        self.knots_per_bs_y_pad = np.asfortranarray(knots_per_bs_y)

    def set_coefs(
        self,
        coefs=None,
        cropbs_neg_flat=None,
    ):

        # ------------
        # check inputs

        if coefs is None:
            msg = "Please provide coefs!"
            raise Exception(msg)

        shape = (self.__nbs[0]*self.__nbs[1],)
        if np.isscalar(coefs):
            self.tck[2][...] = coefs
        else:
            if not isinstance(coefs, np.ndarray):
                coefs = np.asarray(coefs)
            if coefs.ndim > 1 and coefs.size == shape[0]:
                coefs = coefs.ravel()
            if coefs.shape != shape:
                msg = (
                    "Arg coefs has wrong shape!\n"
                    f"\t- expected: {shape}\n"
                    f"\t- provided: {coefs.shape}\n"
                )
                raise Exception(msg)
            self.tck[2][...] = coefs

        # ------------
        # crop and set

        if cropbs_neg_flat is not None:
            self.tck[2][cropbs_neg_flat] = 0.

    def __call__(
        self,
        x,
        y,
        coefs=None,
        cropbs_neg_flat=None,
        **kwdargs,
    ):

        # prepare
        self.set_coefs(
            coefs=coefs,
            cropbs_neg_flat=cropbs_neg_flat,
        )

        # compute
        val = super().__call__(x, y, **kwdargs)

        # clean
        indout = (
            (x < self.tck[0][0]) | (x > self.tck[0][-1])
            | (y < self.tck[1][0]) | (y > self.tck[1][-1])
        )
        val[indout] = np.nan
        return val

    def ev_details(
        self,
        x,
        y,
        indbs_tuple_flat=None,
        reshape=None,
    ):

        # -----------
        # check input

        c0 = (
            isinstance(indbs_tuple_flat, tuple)
            and len(indbs_tuple_flat) == 2
            and all([isinstance(ind, np.ndarray) for ind in indbs_tuple_flat])
            and indbs_tuple_flat[0].shape == indbs_tuple_flat[1].shape
        )
        if not c0:
            msg = (
                "Arg indbs_tuple_flat must be a tuple of indices!"
            )
            raise Exception(msg)

        if reshape is None:
            reshape = True
        if not isinstance(reshape, bool):
            msg = f"Arg reshape must be a bool!\nProvided {reshape}"
            raise Exception(msg)

        # -----------
        # prepare

        deg = self.degrees[0]
        nbs = indbs_tuple_flat[0].size
        shape = x.shape
        x = np.ascontiguousarray(x.ravel(), dtype=np.floating)
        y = np.ascontiguousarray(y.ravel(), dtype=np.floating)
        coef = np.zeros((deg + 4, 1), dtype=float)
        coef[deg] = 1.
        outy = np.full((x.size, 1), np.nan)

        # -----------
        # compute

        val = np.zeros(tuple(np.r_[x.size, nbs]))
        indtot = np.arange(0, nbs)

        iz_u = np.unique(indbs_tuple_flat[1])

        for iz in iz_u:

            scpinterp._bspl.evaluate_spline(
                self.knots_per_bs_y_pad[:, iz],
                coef,
                self.degrees[1],
                y,
                0,
                False,
                outy,
            )

            indoky = ~np.isnan(outy)
            if not np.any(indoky):
                continue
            indokx = np.copy(indoky)

            indr = indbs_tuple_flat[1] == iz
            ir = indbs_tuple_flat[0][indr]
            for ii, iir in enumerate(ir):

                if ii > 0:
                    indokx[...] = indoky

                outx = np.full((indoky.sum(), 1), np.nan)

                scpinterp._bspl.evaluate_spline(
                    self.knots_per_bs_x_pad[:, iir],
                    coef,
                    self.degrees[0],
                    x[indoky[:, 0]],
                    0,
                    False,
                    outx,
                )

                ixok = ~np.isnan(outx)
                if not np.any(ixok):
                    continue

                indokx[indoky] = ixok[:, 0]
                val[indokx[:, 0], indtot[indr][ii]] = (outx[ixok]*outy[indokx])

        if reshape:
            val = np.reshape(val, shape)

        return val

    def get_overlap(self):
        return _get_overlap(
            deg=self.degrees[0],
            knotsx=self.knots_per_bs_x,
            knotsy=self.knots_per_bs_y,
            shapebs=self.shapebs,
        )

    def get_operator(
        self,
        operator=None,
        geometry=None,
        cropbs_flat=None,
    ):
        """ Get desired operator """
        return _mesh_bsplines_operators.get_mesh2dRect_operators(
            deg=self.degrees[0],
            operator=operator,
            geometry=geometry,
            knotsx_mult=self.tck[0],
            knotsy_mult=self.tck[1],
            knotsx_per_bs=self.knots_per_bs_x,
            knotsy_per_bs=self.knots_per_bs_y,
            overlap=self.get_overlap(),
            cropbs_flat=cropbs_flat,
        )


# #############################################################################
# #############################################################################
#                       Mesh2DRect - bsplines - overlap
# #############################################################################


def _get_overlap(
    deg=None,
    knotsx=None,
    knotsy=None,
    shapebs=None,
):
    # nb of overlapping, inc. itself in 1d
    nbsR, nbsZ = shapebs
    indR0 = np.tile(np.arange(0, nbsR), nbsZ)
    indZ0 = np.repeat(np.arange(0, nbsZ), nbsR)

    # complete
    ntot = 2*deg + 1

    addR = np.tile(np.arange(-deg, deg+1), ntot)
    addZ = np.repeat(np.arange(-deg, deg+1), ntot)

    interR = indR0[None, :] + addR[:, None]
    interZ = indZ0[None, :] + addZ[:, None]

    # purge
    inter = interR + interZ*nbsR
    indneg = (
        (interR < 0) | (interR >= nbsR) | (interZ < 0) | (interZ >= nbsZ)
    )
    inter[indneg] = -1

    return inter


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def _get_bs2d_func_check(
    coefs=None,
    ii=None,
    jj=None,
    indbs=None,
    R=None,
    Z=None,
    shapebs=None,
    crop=None,
    cropbs=None,
):

    # ii, jj
    if ii is not None:
        c0 = (
            isinstance(ii, (int, np.integer))
            and ii >= 0
            and ii < shapebs[0]
        )
        if not c0:
            msg = (
                "Arg ii must be an index in the range [0, {shapebs[0]}[\n"
                f"Provided: {ii}"
            )
            raise Exception(msg)
    if jj is not None:
        c0 = (
            isinstance(jj, (int, np.integer))
            and jj >= 0
            and jj < shapebs[1]
        )
        if not c0:
            msg = (
                "Arg jj must be an index in the range [0, {shapebs[1]}[\n"
                f"Provided: {jj}"
            )
            raise Exception(msg)

    # R, Z
    if not isinstance(R, np.ndarray):
        R = np.atleast_1d(R)
    if not isinstance(Z, np.ndarray):
        Z = np.atleast_1d(Z)
    assert R.shape == Z.shape

    # coefs
    if coefs is None:
        coefs = 1.
    if np.isscalar(coefs):
        pass
    else:
        coefs = np.atleast_1d(coefs)
        if coefs.ndim < len(shapebs) or coefs.ndim > len(shapebs) + 1:
            msg = (
                "coefs has too small / big shape!\n"
                f"\t- coefs.shape: {coefs.shape}\n"
                f"\t- shapebs:     {shapebs}"
            )
            raise Exception(msg)
        if coefs.ndim == len(shapebs):
            coefs = coefs.reshape(tuple(np.r_[1, coefs.shape]))
        if coefs.shape[1:] != shapebs:
            msg = (
                "coefs has wrong shape!\n"
                f"\t- coefs.shape: {coefs.shape}\n"
                f"\t- shapebs:     {shapebs}"
            )
            raise Exception(msg)

    # crop
    if crop is None:
        crop = True
    if not isinstance(crop, bool):
        msg = (
            "Arg crop must be a bool!\n"
            f"Provided: {crop}"
        )
        raise Exception(msg)
    crop = crop and cropbs is not None and cropbs is not False

    return coefs, ii, jj, R, Z, crop


def _get_bs2d_func_knots(knots, deg=None, returnas=None, return_unique=None):

    # ----------
    # check input

    returnas = _generic_check._check_var(
        returnas, 'returnas',
        types=str,
        default='data',
        allowed=['ind', 'data'],
    )
    return_unique = _generic_check._check_var(
        return_unique, 'return_unique',
        types=bool,
        default=False,
    )

    # ----------
    # compute

    nkpbs = 2 + deg
    size = knots.size
    nbs = size - 1 + deg

    if return_unique:
        if deg == 0:
            knots_per_bs = np.arange(0, size)
        elif deg == 1:
            knots_per_bs = np.r_[0, np.arange(0, size), size-1]
        elif deg == 2:
            knots_per_bs = np.r_[0, 0, np.arange(0, size), size-1, size-1]
        elif deg == 3:
            knots_per_bs = np.r_[
                0, 0, 0, np.arange(0, size), size-1, size-1, size-1,
            ]

    else:
        knots_per_bs = np.zeros((nkpbs, nbs), dtype=int)

        if deg == 0:
            knots_per_bs[:, :] = np.array([
                np.arange(0, size-1),
                np.arange(1, size),
            ])

        elif deg == 1:
            knots_per_bs[:, 1:-1] = np.array([
                np.arange(0, size-2),
                np.arange(1, size-1),
                np.arange(2, size),
            ])
            knots_per_bs[:, 0] = [0, 0, 1]
            knots_per_bs[:, -1] = [-2, -1, -1]

        elif deg == 2:
            knots_per_bs[:, 2:-2] = np.array([
                np.arange(0, size-3),
                np.arange(1, size-2),
                np.arange(2, size-1),
                np.arange(3, size),
            ])
            knots_per_bs[:, 0] = [0, 0, 0, 1]
            knots_per_bs[:, 1] = [0, 0, 1, 2]
            knots_per_bs[:, -2] = [-3, -2, -1, -1]
            knots_per_bs[:, -1] = [-2, -1, -1, -1]

        elif deg == 3:
            knots_per_bs[:, 3:-3] = np.array([
                np.arange(0, size-4),
                np.arange(1, size-3),
                np.arange(2, size-2),
                np.arange(3, size-1),
                np.arange(4, size),
            ])
            knots_per_bs[:, 0] = [0, 0, 0, 0, 1]
            knots_per_bs[:, 1] = [0, 0, 0, 1, 2]
            knots_per_bs[:, 2] = [0, 0, 1, 2, 3]
            knots_per_bs[:, -3] = [-4, -3, -2, -1, -1]
            knots_per_bs[:, -2] = [-3, -2, -1, -1, -1]
            knots_per_bs[:, -1] = [-2, -1, -1, -1, -1]

    # ----------
    # return

    if returnas == 'data':
        knots_per_bs = knots[knots_per_bs]

    if return_unique:
        return knots_per_bs, nbs
    else:
        return knots_per_bs


def _get_bs2d_func_cents(cents, deg=None, returnas=None):

    returnas = _generic_check._check_var(
        returnas, 'returnas',
        types=str,
        default='data',
        allowed=['ind', 'data'],
    )

    nkpbs = 1 + deg
    size = cents.size
    nbs = size + deg
    cents_per_bs = np.zeros((nkpbs, nbs), dtype=int)

    if deg == 0:
        cents_per_bs[0, :] = np.arange(0, size)

    elif deg == 1:
        cents_per_bs[:, 1:-1] = np.array([
            np.arange(0, size-1),
            np.arange(1, size),
        ])
        cents_per_bs[:, 0] = [0, 0]
        cents_per_bs[:, -1] = [-1, -1]

    elif deg == 2:
        cents_per_bs[:, 2:-2] = np.array([
            np.arange(0, size-2),
            np.arange(1, size-1),
            np.arange(2, size),
        ])
        cents_per_bs[:, 0] = [0, 0, 0]
        cents_per_bs[:, 1] = [0, 0, 1]
        cents_per_bs[:, -2] = [-2, -1, -1]
        cents_per_bs[:, -1] = [-1, -1, -1]

    elif deg == 3:
        cents_per_bs[:, 3:-3] = np.array([
            np.arange(0, size-3),
            np.arange(1, size-2),
            np.arange(2, size-1),
            np.arange(3, size),
        ])
        cents_per_bs[:, 0] = [0, 0, 0, 0]
        cents_per_bs[:, 1] = [0, 0, 0, 1]
        cents_per_bs[:, 2] = [0, 0, 1, 2]
        cents_per_bs[:, -3] = [-3, -2, -1, -1]
        cents_per_bs[:, -2] = [-2, -1, -1, -1]
        cents_per_bs[:, -1] = [-1, -1, -1, -1]

    if returnas == 'data':
        cents_per_bs = cents[cents_per_bs]

    return cents_per_bs


def _get_bs2d_func_max(Rknots=None, Zknots=None, deg=None):

    knots_per_bs_R = _get_bs2d_func_knots(Rknots, deg=deg)
    knots_per_bs_Z = _get_bs2d_func_knots(Zknots, deg=deg)
    nbkbs = knots_per_bs_R.shape[0]

    if nbkbs % 2 == 0:
        ii = int(nbkbs/2)
        Rbs_cent = np.mean(knots_per_bs_R[ii-1:ii+1, :], axis=0)
        Zbs_cent = np.mean(knots_per_bs_Z[ii-1:ii+1, :], axis=0)
        if deg == 2:
            Rbs_cent[:deg] = [Rknots[0], 0.5*(Rknots[0] + Rknots[1])]
            Rbs_cent[-deg:] = [0.5*(Rknots[-2] + Rknots[-1]), Rknots[-1]]
            Zbs_cent[:deg] = [Zknots[0], 0.5*(Zknots[0] + Zknots[1])]
            Zbs_cent[-deg:] = [0.5*(Zknots[-2] + Zknots[-1]), Zknots[-1]]

    else:
        ii = int((nbkbs-1)/2)
        Rbs_cent = knots_per_bs_R[ii, :]
        Zbs_cent = knots_per_bs_Z[ii, :]
        if deg == 1:
            Rbs_cent[:deg] = Rknots[0]
            Rbs_cent[-deg:] = Rknots[-1]
            Zbs_cent[:deg] = Zknots[0]
            Zbs_cent[-deg:] = Zknots[-1]
        elif deg == 3:
            Rbs_cent[:deg] = [Rknots[0], 0.5*(Rknots[0]+Rknots[1]), Rknots[1]]
            Rbs_cent[-deg:] = [
                Rknots[-2], 0.5*(Rknots[-2]+Rknots[-1]), Rknots[-1],
            ]
            Zbs_cent[:deg] = [Zknots[0], 0.5*(Zknots[0]+Zknots[1]), Zknots[1]]
            Zbs_cent[-deg:] = [
                Zknots[-2], 0.5*(Zknots[-2]+Zknots[-1]), Zknots[-1],
            ]

    return Rbs_cent, Zbs_cent


def get_bs2d_RZ(deg=None, Rknots=None, Zknots=None):

    # ----------------
    # get knots per bspline, nb of bsplines...

    knots_per_bs_R = _get_bs2d_func_knots(Rknots, deg=deg, returnas='data')
    knots_per_bs_Z = _get_bs2d_func_knots(Zknots, deg=deg, returnas='data')
    nbkbs = knots_per_bs_R.shape[0]
    shapebs = (knots_per_bs_R.shape[1], knots_per_bs_Z.shape[1])

    # ----------------
    # get centers of bsplines

    Rbs_cent, Zbs_cent = _get_bs2d_func_max(
        Rknots=Rknots, Zknots=Zknots, deg=deg,
    )
    return shapebs, Rbs_cent, Zbs_cent, knots_per_bs_R, knots_per_bs_Z


def get_bs2d_func(
    deg=None,
    Rknots=None,
    Zknots=None,
    shapebs=None,
    knots_per_bs_R=None,
    knots_per_bs_Z=None,
):

    # ----------------
    # Pre-compute bsplines basis elements

    lbr = [
        scpinterp.BSpline.basis_element(
            knots_per_bs_R[:, ii],
            extrapolate=False,
        )
        for ii in range(shapebs[0])
    ]
    lbz = [
        scpinterp.BSpline.basis_element(
            knots_per_bs_Z[:, jj],
            extrapolate=False,
        )
        for jj in range(shapebs[1])
    ]

    RectBiv = [[None for jj in range(shapebs[1])] for ii in range(shapebs[0])]
    for ii in range(shapebs[0]):
        for jj in range(shapebs[1]):
            def func(rr, zz, coefs=None, br=lbr[ii], bz=lbz[jj]):
                if hasattr(coefs, '__iter__'):
                    if rr.ndim == 1:
                        val = coefs[:, None] * (br(rr)*bz(zz))[None, ...]
                    else:
                        val = coefs[:, None, None] * (br(rr)*bz(zz))[None, ...]
                else:
                    val = coefs * br(rr)*bz(zz)
                return val
            RectBiv[ii][jj] = func

    # ----------------
    # Define functions

    RectBiv_scipy = BivariateSplineRect(
        knotsR=Rknots,
        knotsZ=Zknots,
        deg=deg,
        shapebs=shapebs,
    )

    def RectBiv_details(
        R,
        Z,
        shapebs=shapebs,
        RectBiv_scipy=RectBiv_scipy,
        indbs_tuple_flat=None,
        crop=None,
        cropbs=None,
        coefs=None,
        reshape=None,
    ):
        """ Return the value for each point summed on all bsplines """

        # check inputs
        _, _, _, r, z, crop = _get_bs2d_func_check(
            R=R,
            Z=Z,
            shapebs=shapebs,
        )

        # compute
        return RectBiv_scipy.ev_details(
            r,
            z,
            indbs_tuple_flat=indbs_tuple_flat,
            reshape=reshape,
        )

    def RectBiv_sum(
        R,
        Z,
        coefs=None,
        shapebs=shapebs,
        RectBiv=RectBiv,
        crop=None,
        cropbs=None,
        RectBiv_scipy=RectBiv_scipy,
        indbs_tuple_flat=None,
        reshape=None,
    ):
        """ Return the value for each point summed on all bsplines """

        # check inputs
        coefs, _, _, r, z, crop = _get_bs2d_func_check(
            coefs=coefs,
            R=R,
            Z=Z,
            shapebs=shapebs,
            crop=crop,
            cropbs=cropbs,
        )

        # prepare
        nt = 1 if np.isscalar(coefs) else coefs.shape[0]
        shapepts = r.shape

        shape = tuple(np.r_[nt, shapepts])
        val = np.zeros(shape, dtype=float)
        cropbs_neg_flat = ~cropbs.ravel() if crop else None

        # compute
        if np.isscalar(coefs):
            val[0, ...] = RectBiv_scipy(
                r,
                z,
                grid=False,
                coefs=coefs,
                cropbs_neg_flat=cropbs_neg_flat,
            )

        else:
            for ii in range(coefs.shape[0]):
                val[ii, ...] = RectBiv_scipy(
                    r,
                    z,
                    grid=False,
                    coefs=coefs[ii, ...],
                    cropbs_neg_flat=cropbs_neg_flat,
                )

        return val

    return RectBiv_details, RectBiv_sum, RectBiv_scipy
